"""
Base class for Device Mixins
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~DeviceMixinBase
"""

from ophyd import Device


class DeviceMixinBase(Device):
    """
    Base class for apstools Device mixin classes

    .. index:: Ophyd Device Mixin; DeviceMixinBase
    """

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
